(* File: search.ml *)
(* Author: Chad E Brown *)
(* Created: October 2010 *)

open Flags
open State
open Syntax
open Match
open Minisatinterface

let minisatsearchcounter = ref 0;;
let minisatsearchperiod = ref 1;;

let enable_pattern_clauses = ref false;;
let dynamic_pattern_clauses = ref false;;

let minisat_search_period () =
  incr minisatsearchcounter;
  if (!minisatsearchcounter >= !minisatsearchperiod) then
    begin
      minisatsearchcounter := 0;
      if (!verbosity > 3) then begin Printf.printf "About to call minisat_search\n"; flush stdout end;
      let r = minisat_search () in
      if ((!verbosity > 3) && r) then begin Printf.printf "still satisfiable\n"; flush stdout end;
      r
    end
  else
    true

(*** Compute minimal unsatisfiable core with picomus - This should be moved somewhere more reasonable. - Chad, Nov 9, 2010 ***)
let minimal_unsatisfiable_core clauses =
  if (!verbosity > 4) then (Printf.printf "About to call picomus %s\n" (!Config.picomus); flush stdout);
  let (myout,myin,myerr) = Unix.open_process_full ((!Config.picomus) ^ " - -") [| |] in
  Printf.fprintf myin "p cnf %d %d\n" (max_atom ()) (List.length clauses);
  if (!verbosity > 4) then (Printf.printf "About to send clauses to picomus\n"; flush stdout);
  List.iter
    (fun c ->
      Printf.fprintf myin "%s 0\n" (String.concat " " (List.map string_of_int c))
    )
    clauses;
  close_out myin;
  if (!verbosity > 4) then (Printf.printf "About to read result from picomus\n"; flush stdout);
  match (input_line myout) with
  | "s UNSATISFIABLE" ->
      let ret = ref [] in
      let cl = ref [] in
      let neg = ref false in
      let atm = ref 0 in
      begin
	ignore (input_line myout);
	try
	  (*** This is not the best way to do this.  I'll read in one character at a time, building lit ints, clauses (cl), and pushing clauses onto ret to return ***)
	  while true do
	    let c = input_byte myout in (*** will throw End_of_file when finished ***)
	    if (c = 48) then
	      begin
		if ((!atm) = 0) then (** end of clause **)
		  begin
		    ret := (List.rev (!cl))::!ret;
		    cl := []
		  end
		else
		  atm := 10 * (!atm)
	      end
	    else if ((c > 48) && (c <= 57)) then
	      atm := 10 * (!atm) + (c - 48)
	    else if (c = 45) then (*** negative literal ***)
	      neg := true
	    else (*** otherwise assume this is a blank space ***)
	      begin
		if ((!atm) > 0) then
		  begin
		    if (!neg) then
		      cl := ((-(!atm))::(!cl))
		    else
		      cl := ((!atm)::(!cl));
		    neg := false;
		    atm := 0;
		  end
	      end
	  done;
	  raise End_of_file (*** Unreachable ***)
	with
	| End_of_file ->
	    ignore (Unix.close_process_full (myout,myin,myerr));
	    if (!verbosity > 4) then
	      begin
		Printf.printf "Minimized Clause Set\n";
		List.iter
		  (fun cl -> List.iter (fun l -> Printf.printf "%d " l) cl; Printf.printf "\n")
		  (!ret)
	      end;
	    (!ret)
      end
  | _ -> raise (GenericError "strange return from picomus") (*** In case this happens, maybe I should just return all the clauses ***)

let print_subproof_info minclauses cri =
  begin
    if (!verbosity > 0) then
      begin
	List.iter
	  (fun c ->
	    try
	      match (Hashtbl.find cri c) with
	      | InstRule(a,m,n) ->
		  begin
(***
		    match a with
		    | Ar(_,_) -> Printf.printf "HO Instantiation of type %s for\n%s:\n* %s\n" (stp_str a) (trm_str m) (trm_str n)
		    | _ -> () (*** Printf.printf "FO Instantiation of type %s for\n%s:\n* %s\n" (stp_str a) (trm_str m) (trm_str n) ***)
***)
		  end
	      | _ -> ()
	    with
	    | Not_found -> ())
	  minclauses;
      end
  end
    
let rec new_assumption_clause c =
  try
    begin
      insert_assumption_clause c;
      match c with
      |	[] -> raise (Unsatisfiable None)
      |	[l1] ->
	  begin
	    if (!verbosity > 3) then begin Printf.printf "About to call minisat_addClause1 a (%d)\n" l1; flush stdout end;
	    if (minisat_addClause1(l1)) then
	      begin
		if (not (minisat_search_period())) then
		  raise (Unsatisfiable None);
	      end
	    else raise (Unsatisfiable None)
	  end
      |	_ ->
	  begin
	    if (!verbosity > 8) then begin Printf.printf "Adding lits to minisat\n"; flush stdout end;
	    List.iter (fun l -> minisat_addLit l) c;
	    if (!verbosity > 3) then begin Printf.printf "About to call minisat_addClause (%s)\n" (List.fold_left (fun x y -> if ((String.length x) == 0) then (string_of_int y) else (x ^ " " ^ (string_of_int y))) "" c); flush stdout end;
	    if (minisat_addClause()) then
	      begin
		if (not (minisat_search_period())) then
		  raise (Unsatisfiable None)
	      end
	    else raise (Unsatisfiable None)
	  end
    end
  with
  | DuplicateClause -> ()
  | (Unsatisfiable _) ->
      if (!verbosity > 20) then (Printf.printf "Proof found with assumption clauses only (no search)...%d clauses\n" (List.length (!clauses)); flush stdout);
      if (mkprooftermp ()) then
	raise (Unsatisfiable (Some (fun () -> AssumptionConflictR)))
      else
	raise (Unsatisfiable None)

let filterp lit =
  begin
    if (!verbosity > 8) then Printf.printf "trying to filter %d\n" lit;
    if (minisat_search1 lit) then
      begin
	if (!verbosity > 8) then Printf.printf "Not filtering %d\n" lit;
	false
      end
    else
      begin
	if (!verbosity > 8) then Printf.printf "Filtering %d\n" lit;
	true
      end
  end

let filteredp lit = Hashtbl.mem filtered lit

let filter lit =
  if (not (filteredp lit)) then
    if (filterp lit) then
      Hashtbl.add filtered lit () (*** The literal cannot be true. ***)

let filterneg lit = filter (- lit)

let rec new_search_clause c r =
  try
    begin
      insert_search_clause c r;
      match c with
      |	[] -> raise (Unsatisfiable None)
      |	[l1] ->
	  begin
	    if (!verbosity > 3) then begin Printf.printf "About to call minisat_addClause1 b (%d)\n" l1; flush stdout end;
	    if (minisat_addClause1(l1)) then
	      begin
		if (not (minisat_search_period())) then
		  raise (Unsatisfiable None)
	      end
	    else raise (Unsatisfiable None)
	  end
      |	[l1;l2] ->
	  begin
	    if (!verbosity > 3) then begin Printf.printf "About to call minisat_addClause2 (%d %d)\n" l1 l2; flush stdout end;
	    if (minisat_addClause2 l1 l2) then
	      begin
		if (not (minisat_search_period())) then
		  raise (Unsatisfiable None)
	      end
	    else raise (Unsatisfiable None)
	  end
      |	[l1;l2;l3] ->
	  begin
	    if (!verbosity > 3) then begin Printf.printf "About to call minisat_addClause3 (%d %d %d)\n" l1 l2 l3; flush stdout end;
	    if (minisat_addClause3 l1 l2 l3) then
	      begin
		if (not (minisat_search_period())) then
		  raise (Unsatisfiable None)
	      end
	    else raise (Unsatisfiable None)
	  end
      |	_ ->
	  begin
	    if (!verbosity > 8) then begin Printf.printf "Adding lits to minisat\n"; flush stdout end;
	    List.iter (fun l -> minisat_addLit l) c;
	    if (!verbosity > 8) then begin Printf.printf "Adding clause to minisat\n"; flush stdout end;
	    if (!verbosity > 3) then begin Printf.printf "About to call minisat_addClause (%s)\n" (List.fold_left (fun x y -> if ((String.length x) == 0) then (string_of_int y) else (x ^ " " ^ (string_of_int y))) "" c); flush stdout end;
	    if (minisat_addClause()) then
	      begin
		if (not (minisat_search_period())) then
		  raise (Unsatisfiable None)
	      end
	    else raise (Unsatisfiable None)
	  end
    end
  with
  | DuplicateClause -> ()
  | (Unsatisfiable _) ->
      if (!verbosity > 3) then (Printf.printf "Proof found for a goal...%d clauses\n" (List.length (!clauses)); flush stdout);
      if (mkprooftermp ()) then
	let cri = !clause_ruleinfo in (*** remember the clauses and this hashtable for them ***)
(*** To test with minimal unsatisfiable core: ***)
	let minclauses = minimal_unsatisfiable_core (!clauses) in
	if (!verbosity > 3) then (Printf.printf "Reduced to %d clauses\n" (List.length minclauses); flush stdout);
(*** To test all clauses: ***)
(***	    let minclauses = !clauses in ***)
	print_subproof_info minclauses cri;
        if (!verbosity > 0) then (*** Print some statistics about the proof ***)
          begin
            let numassumptionclauses = ref 0 in
            let numsearchclauses = ref 0 in
            let assumptionlits = ref [] in
            let searchlits = Hashtbl.create 100 in
            List.iter
              (fun c ->
                 begin
                   try
                     ignore (Hashtbl.find (!clause_ruleinfo) c);
                     incr numsearchclauses;
                     List.iter (fun x -> if (not (Hashtbl.mem searchlits (abs x))) then Hashtbl.add searchlits (abs x) ()) c
                   with
                   | Not_found ->
                     begin
                       incr numassumptionclauses;
                       match c with
                       | [x] -> assumptionlits := ((abs x)::(!assumptionlits))
                       | _ -> raise (GenericError "Assumption clause is not a unit clause?")
                     end
                 end)
              minclauses;
            List.iter (fun x -> Hashtbl.remove searchlits x) (!assumptionlits);
	    if (!verbosity > 3) then
	      begin
		Printf.printf "Refutation Statistics:\n";
		Printf.printf "Number of assumptions %d\n" (!numassumptionclauses);
		Printf.printf "Number of search clauses %d\n" (!numsearchclauses);
		Printf.printf "Number of search literals %d\n" (Hashtbl.length searchlits);
	      end;
          end;
	raise (Unsatisfiable (Some (fun () -> SearchR (minclauses,Hashtbl.find cri))))
      else
	raise (Unsatisfiable None)

(*** (int of nalllit,locevars,unmatched,abslits,dlist,strictlits,rew,ml,pl,il,cg)
  nalllit : negated literal of universally quantified formula from which I've made this clause.
  locevars : evars local to this clause
  unmatched : metawffs that are not yet ground
  abslits : literals of grounded metawffs
  dlist : disagreement pairs
  strictlits : (optional) list of literals which contained a strict occurence of every metavar in the original clause
  rew : If equational lits have been used to rewrite, this contains the information about the rewrite step: (bool [right to left?],stp [type of eqn],metatrm list)
  ml : list of metatrms that will be used to construct a proof if the clause is used
  pl : predicates for quantified formulas.
  il : instantiations for nonvacuous quantifiers
  cg : function for constructing the propositional clauses once the pattern clause has been fully instantiated.
 ***)
let pattern_clauses : (int * (evar list) * (metatrm list) * (int list) * dpair list * metatrm list option * ((bool * stp * metatrm list) list option) * (metatrm list) * (metatrm list) * (metatrm list) * (trm list -> trm list -> trm list -> trm list -> unit)) list ref
    = ref []
let pattern_clauses_usable : (int * trm) list ref
    = ref []
exception TrivialClause

(*** April 6, 2011 - Chad, I want to dynamically turn these on and off.  The immediate application is to make pattern clauses for transitivity of equation. ***)
let pattern_clauses_forall_as_lit : bool ref = ref false 
let pattern_clauses_onlyallstrict : bool ref = ref false 
let pattern_clauses_eqnlits : bool ref = ref false 

let pattern_clauses_transitivity_types : (stp,unit) Hashtbl.t = Hashtbl.create 10

let not_in_prop_model_delay_p : bool ref = ref false (*** Nov 2011 ***)
let not_in_prop_model_delay : int ref = ref 0        (*** Nov 2011 ***)

let get_literal_fn1 : (ftrma -> int) ref = ref get_literal;;
let get_literal_fn2 : (ftrma -> int) ref = ref get_literal;;

let seen_instantiations : ((int * fhashval),unit) Hashtbl.t = Hashtbl.create 10

let seen_inst_p a ((_,_,_,h) as t) =
  Hashtbl.mem seen_instantiations (a,h)

let see_inst a ((_,_,_,h) as t) =
  Hashtbl.add seen_instantiations (a,h) ()

let reset_search_1 () =
  reset_search();
  Hashtbl.clear seen_instantiations

let rec stp_weight a = 1 (** fake for now **)

let rec tm_weight m = 1 (** fake for now **)

let instantiation_priority a m =
  (get_int_flag "INSTANTIATION_DELAY") + (stp_weight a) + (tm_weight m)

let process_instantiation a m =
  begin (*** Process it now -- making it passive ***)
    if (!verbosity > 5) then (Printf.printf "Instantiating with %s\n" (ftrma_str m));
    add_instantiation a m;
    List.iter
      (fun (lit,p) -> (*** Instantiate known universal predicates ***)
	if (not (filteredp (-lit))) then
	  let pm = mk_ap_norm p m in
	  insertWithPriority (get_int_flag "FORALL_DELAY") (ProcessProp1 pm);
	  let pmlit = !get_literal_fn2 pm in
	  new_search_clause [lit;pmlit]
	    (if (mkprooftermp ()) then (Some(InstRule(a,p,m))) else None)
	    )
      (Hashtbl.find_all univpreds a)
  end

let new_active_instantiation (a:int) m =
  if (get_bool_flag "EAGERLY_PROCESS_INSTANTIATIONS") then (*** Process it now -- making it passive ***)
    process_instantiation a m
  else (*** Determine priority and put it on the queue (active instantiation) ***)
    let d = instantiation_priority a m in
    begin
      if (!verbosity > 20) then (Printf.printf "Delaying with priority %d active instantiation %s\n" d (ftrma_str m));
      insertWithPriority d (NewInst(a,m))
    end

let possibly_new_instantiation a m =
  if (not (seen_inst_p a m)) then
    begin
      see_inst a m;
      new_active_instantiation a m
    end
  else if (!verbosity > 5) then
    begin
      Printf.printf "not really new inst %s\n" (ftrma_str m)
    end

let process_new_name a m d =
  let (sigmal,ar) = argtps_rtp (ty_f_rev a) in
  new_usable_head_rtp ar sigmal m d
(*  iter_term_continuations_rtp ar sigmal m d *) (* to do: update this to use ftrm or find an alternative *)

let process_new_type ar a d =
  new_usable_type_rtp ar a d;
  iter_type_continuations_rtp ar a d

let negimp_rule mlit m m1 m2 =
  let r = if (mkprooftermp ()) then (Some(NegPropRule(m))) else None in
  insertWithPriority (get_int_flag "POST_NOR_L_DELAY") (ProcessProp1 m1);
  insertWithPriority (get_int_flag "POST_NOR_R_DELAY") (ProcessProp1 (fneg m2));
  let m1lit = !get_literal_fn2 m1 in
  let nm2lit = !get_literal_fn2 (fneg m2) in
  new_search_clause [-mlit;m1lit] r;
  new_search_clause [-mlit;nm2lit] r

let imp_rule mlit m m1 m2 =
  let r = if (mkprooftermp ()) then (Some(PosPropRule(m))) else None in
  insertWithPriority (get_int_flag "POST_OR_L_DELAY") (ProcessProp1 (fneg m1));
  insertWithPriority (get_int_flag "POST_OR_R_DELAY") (ProcessProp1 m2);
  let nm1lit = !get_literal_fn2 (fneg m1) in
  let m2lit = !get_literal_fn2 m2 in
  new_search_clause [(-mlit);nm1lit;m2lit] r

let forall_rule nmlit a m1 w =
  if (!verbosity > 5) then begin print_string("Instantiating with " ^ (ftrma_str w)); print_newline(); flush stdout end;
  let m1w = mk_ap_norm m1 w in
  insertWithPriority (get_int_flag "FORALL_DELAY") (ProcessProp1 m1w);
  let m1wlit = !get_literal_fn2 m1w in
  new_search_clause [nmlit;m1wlit]
    (if (mkprooftermp ()) then (Some(InstRule(a,m1,w))) else None)

let negforall_rule mlit a m1 =
  let (ws,w) = get_fresh_name (ty_f_rev a) in
  if (!verbosity > 4) then begin print_string("m1 = " ^ (ftrma_str m1) ^ "\nUsing Fresh Witness " ^ (ftrma_str w)); print_newline(); flush stdout end;
  process_new_name a w (get_int_flag "NEW_HEAD_ENUM_DELAY");
  let m1w = mk_ap_norm m1 w in
  if (!verbosity > 8) then begin print_string(" m1w = " ^ (ftrma_str m1w)); print_newline(); flush stdout end;
  let nm1w = fneg m1w in
  insertWithPriority (get_int_flag "EXISTS_DELAY") (ProcessProp1 nm1w);
  let nm1wlit = !get_literal_fn2 nm1w in
  new_search_clause [-mlit;nm1wlit]
    (if (mkprooftermp ()) then (Some(FreshRule(a,m1,ws))) else None)

let decompose lit a largs rargs r =
  let delay = get_int_flag "POST_DEC_DELAY" in
  let ar = ref a in
  let dlits
    = List.map2
        (fun la ra ->
          match !ar with
          | Ar(a1,a2) ->
	     let deq = mk_neg (mk_eq (ty_f a1) la ra) in
	     insertWithPriority delay (ProcessProp1 deq);
	     !get_literal_fn2 deq
          | _ -> raise (Failure "typing problem in decompose")
        )
        largs rargs
  in
  new_search_clause ((-lit)::dlits) r

let consider_confrontation a neqlit u v eqlit s t =
  let op = Confront(eqlit,neqlit,a,s,t,u,v) in
  let (sh,_) = head_spine_f s in
  let (th,_) = head_spine_f t in
  let (uh,_) = head_spine_f u in
  let (vh,_) = head_spine_f v in
  let delay =
    if (((sh = uh) && (th = vh)) || ((sh = vh) && (th = uh))) then
      get_int_flag "CONFR_SAME2_DELAY"
    else if ((sh = uh) || (th = vh) || (sh = vh) || (th = uh)) then
      get_int_flag "CONFR_SAME1_DELAY"
    else
      get_int_flag "CONFR_DIFF_DELAY"
  in
  insertWithPriority delay op

let confront a eqlit s t neqlit u v =
  if (!verbosity > 9) then Printf.printf "Confront %d %d\n" eqlit neqlit;
  let r = if (mkprooftermp ()) then (Some (ConfrontationRule(eqlit,neqlit))) else None in
  let minuseqlit = (- eqlit) in
  let minusneqlit = (- neqlit) in
  let su = mk_eq a s u in
  let tu = mk_eq a t u in
  let sv = mk_eq a s v in
  let tv = mk_eq a t v in
  let msulit = (- (!get_literal_fn2 su)) in
  let mtulit = (- (!get_literal_fn2 tu)) in
  let msvlit = (- (!get_literal_fn2 sv)) in
  let mtvlit = (- (!get_literal_fn2 tv)) in
  insertWithPriority (get_int_flag "POST_CONFRONT1_DELAY") (ProcessProp1 (mk_neg su));
  insertWithPriority (get_int_flag "POST_CONFRONT2_DELAY") (ProcessProp1 (mk_neg tu));
  insertWithPriority (get_int_flag "POST_CONFRONT3_DELAY") (ProcessProp1 (mk_neg sv));
  insertWithPriority (get_int_flag "POST_CONFRONT4_DELAY") (ProcessProp1 (mk_neg tv));
  new_search_clause [minuseqlit;minusneqlit;msulit;msvlit] r;
  new_search_clause [minuseqlit;minusneqlit;msulit;mtvlit] r;
  new_search_clause [minuseqlit;minusneqlit;mtulit;msvlit] r;
  new_search_clause [minuseqlit;minusneqlit;mtulit;mtvlit] r

let choice_rule ((m,_,_,_) as h) args =
  match (fchoiceop m) with
  | Some(a) ->
      begin
	match args with
	| (pred::_) ->
	   let m1 = mk_ap_norm pred (mk_ap_norm h pred) in
	   let m2 = mk_all a (fneg (mk_ap_norm (uptrm_f pred 1 0) (mk_db 0))) in
	   insertWithPriority (get_int_flag "CHOICE_EMPTY_DELAY") (ProcessProp1 m2);
	   insertWithPriority (get_int_flag "CHOICE_IN_DELAY") (ProcessProp1 m1);
	   new_search_clause [!get_literal_fn2 m1;!get_literal_fn2 m2] (if (mkprooftermp ()) then (Some(ChoiceRule(h,pred))) else None)
	| _ -> raise (GenericError "Choice operator must be applied to at least one argument")
      end
  | None -> ()  

(** this only looks at subterms not below binders now **)
let rec subterms_as_instantiations ((m,_,_,_) as t) =
  match m with
  | F_Ap(m1,m2) ->
      let (ab,h1) = subterms_as_instantiations m1 in
      begin
	match ab with
	| Ar(a,b) -> 
	    begin
	      let (_,h2) = subterms_as_instantiations m2 in
	      if (((Hashtbl.length h1) = 0) && ((Hashtbl.length h2) = 0)) then
		possibly_new_instantiation (ty_f b) t;
	      Hashtbl.iter (fun i _ -> if (not (Hashtbl.mem h1 i)) then Hashtbl.add h1 i ()) h2;
	      (b,h1)
	    end
	| _ -> raise (GenericError ((ftrm_str m) ^ " is ill-typed (in subterms_as_instantiations)"))
      end
  | F_Name(x) ->
     let a = Hashtbl.find name_tp x in
     possibly_new_instantiation (ty_f a) t;
     (a,Hashtbl.create 10)
  | _ -> subterms_as_instantiations_2 t; raise Not_found
and subterms_as_instantiations_2 ((m,_,_,_) as t) =
  match m with
  | F_Imp(m1,m2) ->
     (try ignore (subterms_as_instantiations m1) with Not_found -> ());
     ignore (subterms_as_instantiations m2);
  | F_Eq(_,m1,m2) ->
     (try ignore (subterms_as_instantiations m1) with Not_found -> ());
     ignore (subterms_as_instantiations m2)
  | F_All(a,m1) ->
     ignore (subterms_as_instantiations m1);
  | F_Lam(a,m1) ->
     ignore (subterms_as_instantiations m1);
  | _ -> raise Not_found (* de Bruijns will raise Not_found *)

(*** Enumeration of types can be preserved across subgoals.  The types are global. ***)
let typesgenerated = ref 0  

let typesofdepth : (int,stp) Hashtbl.t = Hashtbl.create 3

let generate_type d a =
  if (!verbosity > 9) then (Printf.printf "Generated type %s of depth %d\n" (stp_str a) d; flush stdout);
  Hashtbl.add typesofdepth d a

let nth_term a n =
  let rec nth_db_aux b i n cx =
    if !verbosity > 9 then (Printf.printf "nth_db_aux in %s %d %d\n" (stp_str b) i n; flush stdout);
    match cx with
    | a::cxr ->
       if n land 1 = 1 then
         let (sigmal,ar) = argtps_rtp a in
         if ar = b then
           (mk_db i,sigmal,n asr 1)
         else
           raise Not_found
       else
         nth_db_aux b (i+1) (n asr 1) cxr
    | _ -> raise Not_found
  in
  let rec nth_name_aux n hl =
    match hl with
    | (sigmal,h,_)::hr ->
       if n land 1 = 1 then
         (h,sigmal,n asr 1)
       else
         nth_name_aux (n asr 1) hr
    | _ -> raise Not_found
  in
  let rec nth_sort_aux n l =
    match l with
    | x::r ->
       let c = n land 1 in
       let n = n asr 1 in
       if c = 0 then
         (Base(x),n)
       else
         nth_sort_aux n r
    | _ -> raise Not_found
  in
  let rec nth_type_aux n =
    let c = n land 1 in
    let n = n asr 1 in
    if c = 0 then
      nth_sort_aux n !name_base_list
    else
      let c = n land 1 in
      let n = n asr 1 in
      if c = 0 then
        (Prop,n)
      else
        let (a1,n) = nth_type_aux n in
        let (a2,n) = nth_type_aux n in
        (Ar(a1,a2),n)
  in
  let rec nth_term_aux a n cx =
    if !verbosity > 9 then (Printf.printf "nth_term_aux in %s %d\n" (stp_str a) n; flush stdout);
    match a with
    | Ar(a1,a2) ->
       let (m,n) = nth_term_aux a2 n (a1::cx) in
       (mk_lam_norm (ty_f a1) m,n)
    | Prop ->
       begin
         let c = n land 3 in
         let n = n asr 2 in
         if c = 0 then
           (mk_fal (),n)
         else if c = 1 then
           let (z,sigmal,n) = nth_db_aux Prop 0 n cx in
           nth_spine_aux z sigmal n cx
         else if c = 2 then
           let (z,sigmal,n) = nth_name_aux n (usable_heads_rtp a) in
           nth_spine_aux z sigmal n cx
         else
           let c = n land 3 in
           let n = n asr 2 in
           if c = 0 then
             let (m1,n) = nth_term_aux Prop n cx in
             let (m2,n) = nth_term_aux Prop n cx in
             (mk_imp m1 m2,n)
           else if c = 1 then
             let (b1,n) = nth_type_aux n in
             let (m1,n) = nth_term_aux Prop n (b1::cx) in
             (mk_all (ty_f b1) m1,n)
           else if c = 2 then
             let (b1,n) = nth_type_aux n in
             let (m1,n) = nth_term_aux b1 n cx in
             let (m2,n) = nth_term_aux b1 n cx in
             (mk_eq (ty_f b1) m1 m2,n)
           else
             if get_bool_flag "CHOICE" then
               let (b1,n) = nth_type_aux n in
               let (sigmal,rtp) = argtps_rtp b1 in
               if rtp = a then
                 nth_spine_aux (mk_choice (ty_f b1)) (Ar(b1,Prop)::sigmal) n cx
               else
                 raise Not_found
             else
               raise Not_found
       end
    | Base(b) ->
       begin
         let c = n land 1 in
         let n = n asr 1 in
         if c = 0 then
           let (z,sigmal,n) = nth_db_aux a 0 n cx in
           nth_spine_aux z sigmal n cx
         else
           let c = n land 1 in
           let n = n asr 1 in
           if n = 0 then
             let (z,sigmal,n) = nth_name_aux n (usable_heads_rtp a) in
             nth_spine_aux z sigmal n cx
           else
             if get_bool_flag "CHOICE" then
               let (b1,n) = nth_type_aux n in
               let (sigmal,rtp) = argtps_rtp b1 in
               if rtp = a then
                 nth_spine_aux (mk_choice (ty_f b1)) (Ar(b1,Prop)::sigmal) n cx
               else
                 raise Not_found
             else
               raise Not_found
       end
  and nth_spine_aux h sigmal n cx =
    if !verbosity > 9 then (Printf.printf "nth_spine_aux in %s %d\n" (ftrma_str h) n; flush stdout);
    match sigmal with
    | [] -> (h,n)
    | sigma::sigmar ->
       let (m,n) = nth_term_aux sigma n cx in
       nth_spine_aux (mk_ap_norm h m) sigmar n cx
  in
  let (m,n) = nth_term_aux a n [] in
  if n = 0 then
    m
  else
    raise Not_found

(*** Assume all types of depth < d are known.  Generate all those of depth d. ***)
let generate_types d =
  if (!verbosity > 9) then (Printf.printf "Generate types called with depth %d\n" d; flush stdout);
  if (d = (get_int_flag "ENUM_O")) then generate_type d Prop;
  if (d = (get_int_flag "ENUM_SORT")) then
    (List.iter (fun a -> generate_type d (Base(a))) !name_base_list);
  let dm = d - 1 - (get_int_flag "ENUM_ARROW") in
  if (dm >= 0) then
    begin
      for i = 0 to dm
      do
	List.iter
	  (fun ai ->
	    List.iter
	      (fun aj ->
		generate_type d (Ar(ai,aj)))
	      (Hashtbl.find_all typesofdepth (dm - i)))
	  (Hashtbl.find_all typesofdepth i)
      done
    end

let rec replace_subterms f ((mm,_,_,_) as m) =
  let n =
    match mm with
     | F_Ap(m1,m2) -> mk_ap_norm (replace_subterms f m1) (replace_subterms f m2)
     | F_Lam(a,m1) -> mk_lam_norm a (replace_subterms f m1)
     | F_Imp(m1,m2) -> mk_imp (replace_subterms f m1) (replace_subterms f m2)
     | F_All(a,m1) -> mk_all a (replace_subterms f m1)
     | F_Eq(a,m1,m2) -> mk_eq a (replace_subterms f m1) (replace_subterms f m2)
     | _ -> m
  in
  try
    f n
  with Not_found -> n

let rec leibeq_to_primeq m =
  let f ((mm,_,_,_) as m) =
    match mm with
    | F_All(a1,(F_Imp((F_Ap((F_DB(0),_,_,_),((_,fv1,_,_) as m1)),_,_,_),(F_Ap((F_DB(0),_,_,_),((_,fv2,_,_) as m2)),_,_,_)),_,_,_)) when bit0_0_30nl fv1 && bit0_0_30nl fv2 ->
       begin
         match ty_f_rev a1 with
         | Ar(a2,Prop) ->
            let meq = mk_eq (ty_f a2) (uptrm_f m1 0 (-1)) (uptrm_f m2 0 (-1)) in
            if !verbosity > 4 then Printf.printf "replacing %s\nwith primitive equation %s\n" (ftrm_str mm) (ftrma_str meq);
            meq
         | _ -> raise Not_found (* shouldn't happen; ill typed *)
       end
    | F_All(a1,(F_Imp((F_All(a2,(F_Ap((F_Ap((F_DB(1),_,_,_),(F_DB(0),_,_,_)),_,_,_),(F_DB(0),_,_,_)),_,_,_)),_,_,_),(F_Ap((F_Ap((F_DB(0),_,_,_),((_,fv1,_,_) as m1)),_,_,_),((_,fv2,_,_) as m2)),_,_,_)),_,_,_)) when bit0_0_30nl fv1 && bit0_0_30nl fv2 ->
       begin
         let meq = mk_eq a2 (uptrm_f m1 0 (-1)) (uptrm_f m2 0 (-1)) in
         if !verbosity > 4 then Printf.printf "replacing %s\nwith primitive equation %s\n" (ftrm_str mm) (ftrma_str meq);
         meq
       end
    | F_All(a1,(F_Imp((F_Ap((F_Ap((F_DB(0),_,_,_),((_,fv1,_,m1h) as m1)),_,_,_),(_,_,_,m2bh)),_,_,_),(F_Ap((F_Ap((F_DB(0),_,_,_),((_,fv2,_,m2h) as m2)),_,_,_),(_,_,_,m1bh)),_,_,_)),_,_,_)) when m1bh = m1h && m2bh = m2h && bit0_0_30nl fv1 && bit0_0_30nl fv2 ->
       begin
         match ty_f_rev a1 with
         | Ar(a2,Ar(a3,Prop)) ->
            let meq = mk_eq (ty_f a2) (uptrm_f m1 0 (-1)) (uptrm_f m2 0 (-1)) in
            if !verbosity > 4 then Printf.printf "replacing %s\nwith primitive equation %s\n" (ftrm_str mm) (ftrma_str meq);
            meq
         | _ -> raise Not_found (* shouldn't happen; ill typed *)
       end
    | _ -> raise Not_found
  in
  replace_subterms f m

exception ExistsToChoiceDone

let rec existstochoice1 ((mm,_,_,_) as m) p =
  match mm with
  | F_All(a,q) ->
      begin
	if p then
          mk_all a (existstochoice1 q p)
	else (*** This is where the action is: Replace (forall x, q[x]) with (q[Eps x:a => ~q[x]]) ***)
          subst_trmtrm_f q 0 (mk_ap_norm (mk_choice a) (mk_lam_norm a (fneg q)))
      end
  | F_Eq(a,m1,m2) ->
     begin
       match ty_f_rev a with
       | Ar(a1,a2) ->
          let m1s = uptrm_f m1 0 1 in
          let m2s = uptrm_f m2 0 1 in
	  existstochoice1
            (mk_all (ty_f a1) (mk_eq (ty_f a2) (mk_ap_norm m1s (mk_db 0)) (mk_ap_norm m2s (mk_db 0))))
	    p
       | _ -> raise ExistsToChoiceDone
     end
  | F_Imp(m1,m2) ->
      begin
	try
	  mk_imp (existstochoice1 m1 (not p)) m2
	with ExistsToChoiceDone
	     -> mk_imp m1 (existstochoice1 m2 p)
      end
  | _ -> raise ExistsToChoiceDone
  
let rec existstochoice m =
  try
    existstochoice (existstochoice1 m true)
  with ExistsToChoiceDone -> m

let already_preprocessed = ref false

let preprocess m =
  let m1 =
    if (get_bool_flag "LEIBEQ_TO_PRIMEQ") then leibeq_to_primeq m else m
  in
  let m2 =
    if (get_bool_flag "EXISTSTOCHOICE") then existstochoice m1 else m1
  in
  m2
                         
let preprocess1 m =
  if (!already_preprocessed) then m else preprocess m

let process_search_option op =
  match op with
  | ProcessProp1((mm,_,_,mh) as m) ->
      begin
	if (not (Hashtbl.mem processed mh)) then
	  begin
	    let mlit = !get_literal_fn2 m in
	    if ((!not_in_prop_model_delay_p) && (!minisatsearchcounter = 0) && ((minisat_modelValue mlit) > 0)) then (*** Chad: Nov 2011 - delay if not true in the current prop approx; Chad: Dec 2012 - added condition that minisatsearchcounter = 0 so that modelValue is only called if the last interaction with minisat actually called search. Otherwise we may get segmentation faults. The effect is that if MINISAT_SEARCH_PERIOD > 1 and NOT_IN_PROP_MODEL_DELAY > 0, then we only check if the prop is in the model when we happen to have called minisat to search. ***)
	      begin
		if (!verbosity > 8) then Printf.printf "using prop model to delay working on %d %s\n" mlit (ftrma_str m);
		insertWithPriority (!not_in_prop_model_delay) op
	      end
	    else
	      begin
		Hashtbl.add processed mh ();
		if (!verbosity > 8) then Printf.printf "working on %d %s\n" mlit (ftrma_str m);
		match mm with
		| F_Imp((nm1,_,_,_) as nm,(F_Fal,_,_,_)) ->
		   begin
                     match nm1 with
                     | F_Fal -> () (*** do nothing ***)
                     | F_Imp(m1,m2) -> negimp_rule mlit nm m1 m2
                     | F_All(a,m1t) ->
                        (* if ((!enable_pattern_clauses) && (get_bool_flag "PATTERN_CLAUSES_FORALL_AS_LIT")) then (apply_pattern_clauses mlit m); *)
                        let m1 = mk_lam_norm a m1t in
		        negforall_rule mlit a m1
                     | F_Eq(a,((_,_,_,m1h) as m1),((_,_,_,m2h) as m2)) ->
			let r = if (mkprooftermp ()) then (Some(NegPropRule(nm))) else None in
			begin
			  if (m1h = m2h) then new_search_clause [-mlit] r; (*** negation of reflexivity ***)
                                              (*			  else if (get_bool_flag "SYM_EQ") then sym_eq_clauses (-mlit) a m1 m2; (*** Equivalence of = with symmetric version. In earlier versions of Satallax (< 2.2) I sent m1=m2 and m2=m1 to the same literal. ***) *)
			  match ty_f_rev a with
			  | Ar(a1,a2) -> (*** FE ***)
                             (***				  let m3 = neg (forall a1 (eq a2 (norm name_def (Ap (shift m1 0 1,DB (0,a1)))) (norm name_def (Ap (shift m2 0 1,DB (0,a1)))))) in  ***)
			     let m3 = mk_neg (mk_all (ty_f a1) (mk_eq (ty_f a2) (mk_ap_norm m1 (mk_db 0)) (mk_ap_norm m2 (mk_db 0)))) in (*** No need to shift, since no dangling DB's [Thanks to Teucke!] ; Mar 15, normalize at the end to handle final eta (see FQ too) ***)
			     insertWithPriority (get_int_flag "POST_FEQ_DELAY") (ProcessProp1 m3);
			     if (get_bool_flag "INSTANTIATE_WITH_FUNC_DISEQN_SIDES") then
			       begin
				 possibly_new_instantiation a m1;
				 possibly_new_instantiation a m2
			       end;
			     let m3lit = !get_literal_fn2 m3 in
			     new_search_clause [-mlit;m3lit] r
			  | Prop -> (*** BE ***)
			     let m1lit = !get_literal_fn2 m1 in
			     let m2lit = !get_literal_fn2 m2 in
			     insertWithPriority (get_int_flag "POST_NEQO_L_DELAY") (ProcessProp1 m1);
			     insertWithPriority (get_int_flag "POST_NEQO_R_DELAY") (ProcessProp1 m2);
			     insertWithPriority (get_int_flag "POST_NEQO_NL_DELAY") (ProcessProp1 (fneg m1));
			     insertWithPriority (get_int_flag "POST_NEQO_NR_DELAY") (ProcessProp1 (fneg m2));
			     new_search_clause [-mlit;m1lit;m2lit] r;
			     new_search_clause [-mlit;-m1lit;-m2lit] r
			  | Base(aname) -> (*** decompose, confront, Choice accessible, special accessible ***)
			     if ((get_bool_flag "FILTER_NEGEQ") && (filterp mlit)) then
			       ()
			     else
			       begin
				 Hashtbl.add neqns aname (mlit,m1,m2);
				 List.iter
				   (fun (otherlit,n1,n2) -> (** (delayed) confrontation **)
				     consider_confrontation a mlit m1 m2 otherlit n1 n2)
				   (Hashtbl.find_all peqns aname);
				 possibly_new_instantiation a m1;
				 possibly_new_instantiation a m2;
				 let ((lh1,_,_,_) as lh,largs) = head_spine_f m1 in
				 let ((rh1,_,_,_) as rh,rargs) = head_spine_f m2 in
				 begin
				   match (lh1,rh1) with
				   | (F_Choice(la),F_Choice(ra)) when (la = ra) ->
                                      let la1 = ty_f_rev la in
				      decompose mlit (Ar(Ar(la1,Prop),la1)) largs rargs r
				   | (F_Name (lhname),F_Name (rhname)) when ((lhname = rhname) && (decomposable lhname)) ->
                                      begin
                                        try
					  decompose mlit (Hashtbl.find name_tp lhname) largs rargs r
                                        with Not_found -> ()
                                      end
				   | _ -> ()
				 end;
				 choice_rule lh largs;
				 choice_rule rh rargs;
			       end
			end
                     | _ ->
		        let ((h1,_,_,_) as h,args) = head_spine_f nm in
                        match h1 with
		        | F_Choice(a) ->
			   if ((get_bool_flag "FILTER_NEGATM") && (filterp mlit)) then
			     ()
			   else
			     begin
			       let pmd = get_int_flag "PRE_CHOICE_MATING_DELAY_NEG" in
			       Hashtbl.add nchoiceatoms a (mlit,args);
                               let a1 = ty_f_rev a in
                               let a2 = Ar(Ar(a1,Prop),a1) in
			       List.iter (fun (plit,pargs) ->
				   if (not (pargs = args)) then
				     insertWithPriority pmd (Mating(plit,mlit,a2,pargs,args))
				 )
				 (Hashtbl.find_all pchoiceatoms a);
			       choice_rule h args;
			     end
		        | F_Name(x) -> (** negative atom **)
			   if ((get_bool_flag "FILTER_NEGATM") && (filterp mlit)) then
			     ()
			   else
			     begin
			       if (decomposable x) then
				 begin
				   let pmd = get_int_flag "PRE_MATING_DELAY_NEG" in
				   Hashtbl.add natoms x (mlit,args);
                                   let a2 = Hashtbl.find name_tp x in
				   List.iter (fun (plit,pargs) ->
				       if (not (pargs = args)) then
				         insertWithPriority pmd (Mating(plit,mlit,a2,pargs,args))
				     )
				     (Hashtbl.find_all patoms x)
				 end;
			       choice_rule h args;
			     end
		        | _ ->
                           Printf.printf "nm: %s\n" (ftrma_str nm);
                           Printf.printf "nm1: %s\n" (ftrm_str nm1);
                           raise (GenericError ("Unhandled case 1 (were logical constants normalized?) h:" ^ (ftrma_str h)))
		   end
		| F_All(a,m1t) -> (*** Instantiate ***)
		    begin
                      (*		      if (!enable_pattern_clauses) then (apply_pattern_clauses mlit m; if (!dynamic_pattern_clauses) then new_pattern_clauses (-mlit) m); *)
		      if ((get_bool_flag "FILTER_UNIV") && (filterp mlit)) then
			()
		      else
			let nmlit = (- mlit) in
			begin
                          let m1 = mk_lam_norm a m1t in
			  Hashtbl.add univpreds a (nmlit,m1);
                          (*                          if (get_bool_flag "HOUNIF1") then hounif1 (!verbosity) (to_meta m) [] [] [] (get_int_flag "HOUNIF1BOUND") true possibly_new_instantiation; (*** Mar 2012 ***) *)
			  let insts = (get_instantiations a) in
			  begin
			    match ty_f_rev a with
			    | Ar(a1,a2) ->
				begin
				  let rta = rtp a2 in
				  begin
				    match rta with
				    | Base(rtaname) ->
					if (not (default_elt_p rtaname)) then
					  (insertWithPriority (get_int_flag "DEFAULTELT_DELAY") (DefaultElt rtaname));
				    | _ -> (*** Prop: No default element ***)
					()
				  end;
                                  begin (** since there is a quantifier of arrow type, start a simple enumeration scheme for instantiations. This is needed for theoretical completeness but is utterly impractical. **)
                                    if not (enum_of_started a) then
                                      let d = get_int_flag "ENUM_START" in
                                      enum_of_start a;
                                      insertWithPriority d (EnumInst(ty_f_rev a,0));
                                  end;
                                  (** omitting old enumeration completely for now; rethink for ftrms **)
                                  (**
				  begin
				    if (get_bool_flag "ENUM_ITER_DEEP") then
				      begin (*** Iterative Deepening Enumeration Scheme ***)
					if (not (enum_of_started a)) then
					  let d = get_int_flag "ENUM_START" in
					  begin
					    if (not (!enum_started)) then
					      let d_const = get_int_flag "IMITATE_DELAY" in
					      let d_def = get_int_flag "IMITATE_DEFN_DELAY" in
					      begin
						enum_started := true;
						List.iter
						  (fun (x,m,sigma) ->
						    let (sigmal,rtp) = argtps_rtp sigma in
						    if (Hashtbl.mem name_def x) then
						      begin
							if (get_bool_flag "IMITATE_DEFNS") then
							  new_usable_head_rtp rtp sigmal m d_def
						      end
						    else
						      new_usable_head_rtp rtp sigmal m d_const
							)
						  !name_trm_list
					      end;
					    enum_of_start a;
					    insertWithPriority d (EnumIterDeep(get_int_flag "ENUM_ITER_DEEP_INIT",a))
					  end
				      end
				    else (*** The Original Enumeration Scheme ***)
				      begin
					begin
				      (** If we haven't started generating instantiations, start now. **)
					  if (not (!enum_started)) then
					    let d = get_int_flag "ENUM_START" in
					    let d_o = get_int_flag "ENUM_O" in
					    let d_i = get_int_flag "ENUM_SORT" in
					    let d_const = get_int_flag "IMITATE_DELAY" in
					    let d_def = get_int_flag "IMITATE_DEFN_DELAY" in
					    begin
					      enum_started := true;
					      insertWithPriority (d + d_o) (EnumTp(d_o,Prop,Prop));
					      List.iter
						(fun beta -> insertWithPriority (d + d_i) (EnumTp(d_i,Base(beta),Base(beta))))
						!name_base_list;
					      List.iter
						(fun (x,m,sigma) ->
						  let (sigmal,rtp) = argtps_rtp sigma in
						  if (Hashtbl.mem name_def x) then
						    begin
						      if (get_bool_flag "IMITATE_DEFNS") then
							new_usable_head_rtp rtp sigmal m d_def
						    end
						  else
						    new_usable_head_rtp rtp sigmal m d_const
						      )
						!name_trm_list
					    end
					end;
					begin
					  if (not (enum_of_started a)) then
					    let d = get_int_flag "ENUM_START" in
					    begin
					      enum_of_start a;
					      insertWithPriority d (Enum(0,[],a,(fun m -> let m = norm name_def m in possibly_new_instantiation a m)))
					    end
					end
				      end
				  end
                                   **)
				end
			    | Prop -> ()
			    | Base(aname) ->
				begin
				  match insts with
				  | [] ->
				      insertWithPriority (get_int_flag "DEFAULTELTINST_DELAY") (DefaultEltInst aname)
				  | _ -> ()
				end
			  end;
			  List.iter
			    (fun w -> forall_rule nmlit a m1 w)
			    insts
			end
		    end
                | F_Fal ->
		   let r = if (mkprooftermp ()) then (Some(PosPropRule(m))) else None in
		   let flit = !get_literal_fn2 (mk_fal()) in
		   new_search_clause [-flit] r
                | F_Imp(m1,m2) ->
		   begin
                     (*		    if (!enable_pattern_clauses) then (apply_pattern_clauses mlit m); *)
                     imp_rule mlit m m1 m2
		   end
                | F_Eq(a,m1,m2) ->
		   let r = if (mkprooftermp ()) then (Some(PosPropRule(m))) else None in
		   begin
                     (*				  if ((not (m1 = m2)) && (get_bool_flag "SYM_EQ")) then sym_eq_clauses mlit a m1 m2; (*** Equivalence of = with symmetric version. In earlier versions of Satallax (< 2.2) I sent m1=m2 and m2=m1 to the same literal. ***) *)
                     (*				  if ((get_bool_flag "PATTERN_CLAUSES_TRANSITIVITY_EQ") && (not (Hashtbl.mem pattern_clauses_transitivity_types a))) then
				                  begin (*** April 6, 2011 - add a pattern clause for transitivity of equality the first time we see an equation at a type. ***)
				                  let transa = forall a (forall a (forall a (imp (eq a (DB(2,a)) (DB(1,a))) (imp (eq a (DB(1,a)) (DB(0,a))) (eq a (DB(2,a)) (DB(0,a))))))) in
				                  let transalit = !get_literal_fn2 transa in
				                  let transa2 = forall a (forall a (forall a (imp (eq a (DB(2,a)) (DB(1,a))) (imp (eq a (DB(0,a)) (DB(1,a))) (eq a (DB(2,a)) (DB(0,a))))))) in
				                  let transa2lit = !get_literal_fn2 transa2 in
				                  let transa3 = forall a (forall a (forall a (imp (eq a (DB(1,a)) (DB(2,a))) (imp (eq a (DB(1,a)) (DB(0,a))) (eq a (DB(2,a)) (DB(0,a))))))) in
				                  let transa3lit = !get_literal_fn2 transa3 in
				                  Hashtbl.add pattern_clauses_transitivity_types a ();
				                  pattern_clauses_forall_as_lit := false;
				                  pattern_clauses_onlyallstrict := false;
				                  new_search_clause [transalit] (if (mkprooftermp ()) then (Some (Known(transalit,coqknown("@eq_trans","eq_trans"),[a]))) else None);
				                  new_search_clause [transa2lit] (if (mkprooftermp ()) then (Some (Known(transa2lit,coqknown("eq_trans2","eq_symtrans1"),[a]))) else None);
				                  new_search_clause [transa3lit] (if (mkprooftermp ()) then (Some (Known(transa3lit,coqknown("eq_trans3","eq_symtrans2"),[a]))) else None);
				                  new_pattern_clauses (- transalit) transa;
				                  new_pattern_clauses (- transa2lit) transa2;
				                  new_pattern_clauses (- transa3lit) transa3;
				                  pattern_clauses_forall_as_lit := get_bool_flag "PATTERN_CLAUSES_FORALL_AS_LIT";
				                  pattern_clauses_onlyallstrict := get_bool_flag "PATTERN_CLAUSES_ONLYALLSTRICT";
				                  end;  *)
		     match ty_f_rev a with
		     | Ar(a1,a2) -> (*** FQ ***)
                        (*			if (!enable_pattern_clauses) then (apply_pattern_clauses mlit m); *)
                        (***				  let m3 = forall a1 (eq a2 (norm name_def (Ap (shift m1 0 1,DB (0,a1)))) (norm name_def (Ap (shift m2 0 1,DB (0,a1))))) in ***)
			let m3 = mk_all (ty_f a1) (mk_eq (ty_f a2) (mk_ap_norm m1 (mk_db 0)) (mk_ap_norm m2 (mk_db 0))) in (*** See FE above for why there's no shift on m1 and m2. Mar 15 2011: Normalize after everything, to make sure the case of a final eta is handled, e.g., forall x.s=x where s has no x free eta reduces. ***)
			insertWithPriority (get_int_flag "POST_NFEQ_DELAY") (ProcessProp1 m3);
			let m3lit = !get_literal_fn2 m3 in
			new_search_clause [-mlit;m3lit] r
		     | Prop -> (*** BQ ***)
                        (*			if (!enable_pattern_clauses) then (apply_pattern_clauses mlit m); *)
			let m1lit = !get_literal_fn2 m1 in
			let m2lit = !get_literal_fn2 m2 in
			insertWithPriority (get_int_flag "POST_EQO_L_DELAY") (ProcessProp1 m1);
			insertWithPriority (get_int_flag "POST_EQO_R_DELAY") (ProcessProp1 m2);
			insertWithPriority (get_int_flag "POST_EQO_NL_DELAY") (ProcessProp1 (fneg m1));
			insertWithPriority (get_int_flag "POST_EQO_NR_DELAY") (ProcessProp1 (fneg m2));
			new_search_clause [-mlit;m1lit;-m2lit] r;
			new_search_clause [-mlit;-m1lit;m2lit] r
		     | Base(aname) -> (*** confront ***)
			if ((get_bool_flag "FILTER_POSEQ") && (filterp mlit)) then
			  ()
			else
			  begin
                            (*			    if (!enable_pattern_clauses) then (apply_pattern_clauses mlit m); *)
			    Hashtbl.add peqns aname (mlit,m1,m2);
			    List.iter
			      (fun (otherlit,n1,n2) -> (*** (delayed) confrontation ***)
				consider_confrontation a otherlit n1 n2 mlit m1 m2)
			      (Hashtbl.find_all neqns aname);
			  end;
		   end
		| _ ->
 		   begin
		     let ((h1,_,_,_) as h,args) = head_spine_f m in
		     match h1 with
		     | F_Choice(a) ->
			let pmd = get_int_flag "PRE_CHOICE_MATING_DELAY_POS" in
			if ((get_bool_flag "FILTER_POSATM") && (filterp mlit)) then
			  ()
			else
			  begin
                            (*			      if (!enable_pattern_clauses) then (apply_pattern_clauses mlit m); *)
			    Hashtbl.add pchoiceatoms a (mlit,args);
                            let a1 = ty_f_rev a in
                            let a2 = Ar(Ar(a1,Prop),a1) in
			    List.iter (fun (nlit,nargs) ->
				if (not (args = nargs)) then
				  insertWithPriority pmd (Mating(mlit,nlit,a2,args,nargs));
			      )
			      (Hashtbl.find_all nchoiceatoms a);
			    choice_rule h args;
			  end
		     | F_Name(x) -> (** positive atom **)
			  if ((get_bool_flag "FILTER_POSATM") && (filterp mlit)) then
			    ()
			  else
			    begin
                              (*			      if (!enable_pattern_clauses) then (apply_pattern_clauses mlit m); *)
			      if (decomposable x) then
				begin
				  let pmd = get_int_flag "PRE_MATING_DELAY_POS" in
				  Hashtbl.add patoms x (mlit,args);
                                  let a2 = Hashtbl.find name_tp x in
				  List.iter (fun (nlit,nargs) ->
				    if (not (args = nargs)) then
				      insertWithPriority pmd (Mating(mlit,nlit,a2,args,nargs));
				    )
				    (Hashtbl.find_all natoms x)
				end;
			      choice_rule h args; (*** This was missing and caused some incompleteness (eg, CHOICE6 in mode59c). - Chad, Jan 25, 2011 ***)
			    end
		      | _ -> raise (GenericError ("Unhandled case 2 (were logical constants normalized?) h:" ^ (ftrma_str h)))
		    end
	      end
	  end (*** no else ***)
      end
  | Mating(plit,nlit,a,pargs,nargs) ->
      if (!verbosity > 9) then Printf.printf "Mating %d %d\n" plit nlit;
      let pmd = get_int_flag "POST_MATING_DELAY" in
      let ar = ref a in
      let dlits
        = List.map2
	    (fun pa na ->
              match !ar with
              | Ar(a,b) ->
	         let deq = mk_neg (mk_eq (ty_f a) pa na) in
                 ar := b;
	         insertWithPriority pmd (ProcessProp1 deq);
	         !get_literal_fn2 deq
              | _ -> raise (GenericError "Mating typing problem"))
	    pargs nargs
      in
      new_search_clause
	((-plit)::(-nlit)::dlits)
	(if (mkprooftermp ()) then (Some (MatingRule(plit,nlit))) else None)
  | Confront(eqlit,neqlit,a,s,t,u,v) ->
      confront a eqlit s t neqlit u v
  | DefaultElt(aname) -> ignore (default_elt aname)
  | DefaultEltInst(aname) -> (*** If there are no instantiations of the sort, use the default elt. ***)
      let a = ty_f (Base aname) in
      let insts = (get_instantiations a) in
      begin
	match insts with
	| [] -> 
	    let m = default_elt aname in
	    possibly_new_instantiation a m
	| _ -> ()
      end
  | NewInst(a,m) ->
     process_instantiation a m; (*** Must process it now and make it passive ***)
  | EnumInst(a,n) -> (* generate the nth term and add it as an instantiation *)
     begin
       try
         let m = nth_term a n in
         insertWithPriority 500 (NewInst(ty_f a,m));
         insertWithPriority 1000 (EnumInst(a,n+1)) (* with low priority generate the next one *)
       with Not_found ->
         insertWithPriority 1000 (EnumInst(a,n+1)) (* with low priority generate the next one *)
     end
  | Filter(d) ->
      try
	let (d1,_) = peekAtNext() in (*** peek to make sure there is some other option, use this to delay the next Filter ***)
	(*** Put Filter back on priority queue, with higher delay ( = lower priority) ***)
	let d2 = d1 + (get_int_flag "FILTER_INCR") in
	insertWithPriority d2 (Filter d2);
	if (get_bool_flag "FILTER_UNIV_USABLE") then
	  Hashtbl.iter (fun a (n,_) -> filterneg n)
	    univpreds;
	if (get_bool_flag "FILTER_POSATM_USABLE") then
	  begin
	    Hashtbl.iter (fun x (n,_) -> filterneg n)
	      patoms;
	    Hashtbl.iter (fun x (n,_) -> filterneg n)
	      pchoiceatoms
	  end;
	if (get_bool_flag "FILTER_NEGATM_USABLE") then
	  begin
	    Hashtbl.iter (fun x (n,_) -> filterneg n)
	      natoms;
	    Hashtbl.iter (fun x (n,_) -> filterneg n)
	      nchoiceatoms
	  end;
	if (get_bool_flag "FILTER_POSEQ_USABLE") then
	  Hashtbl.iter (fun x (n,_,_) -> filterneg n)
	    peqns;
	if (get_bool_flag "FILTER_NEGEQ_USABLE") then
	  Hashtbl.iter (fun x (n,_,_) -> filterneg n)
	    neqns;
      with
      | Not_found -> () (*** If this is the only option, quit ***)

let search_main () =
  try
    while true do
      if (!verbosity > 5) then
	begin
	  let (p,op) = peekAtNext() in
	  Printf.printf "Option (Priority %d): %s\n" p (searchoption_str op)
	end;
      let op = getNext() in
      try
	process_search_option op
      with
      | Not_found -> raise (GenericError "Not_found raised while processing search option")
    done
  with Not_found ->
    begin
      match (!mkproofterm) with
      | Some ModelTrue ->
	  print_model true
      | Some Model ->
	  print_model false
      | _ -> ()
    end;
    if (!minisatsearchperiod > 1) then
      begin
	let r = minisat_search () in
	if (not r) then
	  raise (Unsatisfiable None)
      end;
    raise Satisfiable

let pos_heads : (fhashval,unit) Hashtbl.t = Hashtbl.create 10
let neg_heads : (fhashval,unit) Hashtbl.t = Hashtbl.create 10
let eqn_heads : (fhashval,unit) Hashtbl.t = Hashtbl.create 10
let diseqn_heads : (fhashval,unit) Hashtbl.t = Hashtbl.create 10

let rec set_relevance_info_tm ((mm,_,_,_) as m) =
  match mm with
  | F_Lam(a,m1) -> set_relevance_info_tm m1
  | _ ->
      match (head_spine_f m) with
      | ((F_Name(x),_,_,h),s) ->
	 Hashtbl.add diseqn_heads h ();
	 ignore (List.map set_relevance_info_tm s)
      |	_ -> ()

let rec set_relevance_info ((mm,_,_,_) as m) pos =
  match mm with
  | F_Lam(a,m1) -> set_relevance_info m1 pos
  | F_All(a,m1) -> set_relevance_info m1 pos
  | F_Imp(m1,m2) ->
     set_relevance_info m1 (not pos);
     set_relevance_info m2 pos
  | F_Eq(_,m1,m2) ->
     begin
       match (head_spine_f m1) with
       | ((F_Name(x),_,_,h),s) ->
	  if pos then
	    Hashtbl.add eqn_heads h ()
	  else
	    (Hashtbl.add diseqn_heads h ();
	     ignore (List.map set_relevance_info_tm s))
       |  _ -> ();
     end;
     begin
       match (head_spine_f m2) with
       | ((F_Name(x),_,_,h),s) ->
	  if pos then
	    Hashtbl.add eqn_heads h ()
	  else
	    (Hashtbl.add diseqn_heads h ();
	     ignore (List.map set_relevance_info_tm s))
       |  _ -> ()
     end
  | _ ->
      match (head_spine_f m) with
      | ((F_Name(x),_,_,h),s) ->
	  ignore (List.map set_relevance_info_tm s);
	  if pos then
	    Hashtbl.add pos_heads h ()
	  else
	    Hashtbl.add neg_heads h ()
      | (_,_) -> ()

let rec relevance_delay ((mm,_,_,_) as m) pos d =
  if (d <= 0) then
    0
  else
    match mm with
    | F_Lam(a,m1) -> relevance_delay m1 pos d
    | F_All(a,m1) -> relevance_delay m1 pos d
    | F_Imp(m1,m2) -> relevance_delay m1 (not pos) (relevance_delay m2 pos d)
    | F_Eq(a,m1,m2) ->
       begin
	 let d2 =
	   match (head_spine_f m1) with
	   | ((F_Name(x),_,_,h),s) ->
	      if ((pos && (Hashtbl.mem diseqn_heads h)) || ((not pos) && (Hashtbl.mem eqn_heads h))) then
		d - 1
	      else
		d
	   | _ -> d
	 in
	 match (head_spine_f m2) with
	 | ((F_Name(x),_,_,h),s) ->
	    if ((pos && (Hashtbl.mem diseqn_heads h)) || ((not pos) && (Hashtbl.mem eqn_heads h))) then
	      d2 - 1
	    else
	      d2
	 | _ -> d2
       end
    | _ ->
	match (head_spine_f m) with
	| ((F_Name(x),_,_,h),s) ->
	   if ((pos && (Hashtbl.mem neg_heads h)) || ((not pos) && (Hashtbl.mem pos_heads h))) then
	     d - 1
	   else
	     d
	| (_,_) -> d

let search_1 b1 b2 =
  begin
    let _ = minisat_init (!verbosity) in
    if (!verbosity > 3) then Printf.printf "Initialized minisat\n";
    (*** Process "conjecture" part first - March 18 2011 ***)
    List.iter (fun (m:ftrma) ->
      let mn = preprocess1 m in
      match (fchoiceop_axiom mn) with
      | Some(x,a) ->
         if (!verbosity > 3) then Printf.printf "choice: %s %s\n" x (stp_str a);
	  declare_fchoiceop x (ty_f a) mn
      | None ->
	  begin
	    insertWithPriority 0 (ProcessProp1 mn);
	    if ((get_int_flag "RELEVANCE_DELAY") > 0) then
	      set_relevance_info mn false;
	    let mlit = !get_literal_fn1 mn in
	    new_assumption_clause [mlit];
	    if (get_bool_flag "INITIAL_SUBTERMS_AS_INSTANTIATIONS") then
	      begin
		if (!verbosity > 4) then Printf.printf "---- Initial Subterms as Instantiations BEGIN ----\n";
		if (!verbosity > 4) then Printf.printf "---- m = %s ----\n" (ftrma_str m);
		if (!verbosity > 4) then Printf.printf "---- mn = %s ----\n" (ftrma_str mn);
		(try ignore (subterms_as_instantiations mn) with Not_found -> ());
		if (!verbosity > 4) then Printf.printf "---- Initial Subterms as Instantiations END ----\n"
	      end
	  end)
      b1;
    (*** Process "axiom" part second - March 18 2011 ***)
    List.iter (fun m ->
      let mn = preprocess1 m in
      match (fchoiceop_axiom mn) with
      | Some(x,a) ->
         if (!verbosity > 3) then Printf.printf "choice: %s %s\n" x (stp_str a);
	  declare_fchoiceop x (ty_f a) mn
      | None ->
	  begin
	    insertWithPriority
		((get_int_flag "AXIOM_DELAY") + (relevance_delay mn true (get_int_flag "RELEVANCE_DELAY")))
	      (ProcessProp1 mn);
	    let mlit = !get_literal_fn1 mn in
	    new_assumption_clause [mlit];
	    if (get_bool_flag "INITIAL_SUBTERMS_AS_INSTANTIATIONS") then
	      begin
		if (!verbosity > 4) then Printf.printf "---- Initial Subterms as Instantiations BEGIN ----\n";
		if (!verbosity > 4) then Printf.printf "---- m = %s ----\n" (ftrma_str m);
		if (!verbosity > 4) then Printf.printf "---- mn = %s ----\n" (ftrma_str mn);
		(try ignore (subterms_as_instantiations mn) with Not_found -> ());
		if (!verbosity > 4) then Printf.printf "---- Initial Subterms as Instantiations END ----\n"
	      end
	  end)
      b2;
    begin
      if ((get_bool_flag "FILTER_UNIV_USABLE") || (get_bool_flag "FILTER_POSATM_USABLE") || (get_bool_flag "FILTER_NEGATM_USABLE") || (get_bool_flag "FILTER_POSEQ_USABLE") || (get_bool_flag "FILTER_NEGEQ_USABLE")) then
	let d = (get_int_flag "FILTER_START") in
	insertWithPriority d (Filter d)
    end;
    search_main ()
  end

let nontheorem_split_timeout : float ref = ref 16.0 (*** default of 16 seconds ***)

let rec split_global_disjunctions b1 b2 b bs sgn cnj =
  match b1 with
    (((mm,_,_,_) as m)::pr) ->
      begin
	match mm with
        | F_Imp((F_Eq(_,(_,_,_,m1h),(_,_,_,m2h)),_,_,_),(F_Fal,_,_,_)) when m1h = m2h ->
	   if (mkprooftermp ()) then
	     raise (Unsatisfiable (Some (fun () -> NegReflR)))
	   else
	     raise (Unsatisfiable None)
        | F_Imp((F_All(a,m1),_,_,_),(F_Fal,_,_,_)) ->
	    begin
	      let (ws,w) = get_fresh_name (ty_f_rev a) in
	      if (!verbosity > 5) then begin print_string("Using Fresh Witness " ^ (ftrma_str w)); print_newline(); flush stdout end;
	      process_new_name a w (get_int_flag "NEW_HEAD_ENUM_DELAY");
	      let m1w = subst_trmtrm_f m1 0 w in
	      if (!verbosity > 8) then (print_string(" m1w = " ^ (ftrma_str m1w)); print_newline());
	      let nm1w = fneg m1w in
	      if (!verbosity > 8) then (print_string(" nm1w = " ^ (ftrma_str nm1w)); print_newline());
	      begin
		try
		  split_global_disjunctions (nm1w::pr) b2 b bs sgn cnj
		with
		| Unsatisfiable(Some(r1)) ->
		   raise (Unsatisfiable(Some(fun () -> NegAllR(ty_f_rev a,ftrma_trm name_tp [] m1,ws,r1 ()))))
	      end
	    end
	| F_Imp((F_Fal,_,_,_),(F_Fal,_,_,_)) ->
           split_global_disjunctions pr b2 b bs sgn cnj (*** drop ~False ***)
	| F_Imp((F_Imp(m1,m2),_,_,_),(F_Fal,_,_,_)) ->
	   begin
	     try
	       split_global_disjunctions (m1::(fneg m2)::pr) b2 b bs sgn cnj
	     with
	     | Unsatisfiable(Some(r1)) ->
		raise (Unsatisfiable (Some (fun () -> NegImpR(ftrma_trm name_tp [] m1,ftrma_trm name_tp [] m2,r1 ()))))
	   end
	| F_Imp((F_Eq(a,m1,m2),_,_,_),(F_Fal,_,_,_)) ->
           begin
             match ty_f_rev a with
             | Ar(a1,a2) -> (** no need to shift since no dangling de Bruijns *)
		begin
		  try
		    split_global_disjunctions
                      ((mk_neg (mk_all (ty_f a1) (mk_eq (ty_f a2) (mk_ap_norm m1 (mk_db 0)) (mk_ap_norm m2 (mk_db 0)))))::pr)
		      b2 b bs sgn cnj
		  with
		  | Unsatisfiable(Some(r1)) ->
		     raise (Unsatisfiable (Some (fun () -> NegEqFR(a1,a2,ftrma_trm name_tp [] m1,ftrma_trm name_tp [] m2,r1 ()))))
		end
             | Prop ->
		begin
		  try
		    split_global_disjunctions (m1::(fneg m2)::pr) b2 b bs sgn cnj
		  with
		  | Unsatisfiable(Some(r1)) ->
		     begin
		       try
			 reset_search_1 (); incr sgn;
			 split_global_disjunctions ((fneg m1)::m2::pr) b2 b bs sgn cnj
		       with
		       | Unsatisfiable(Some(r2)) ->
			  raise (Unsatisfiable (Some (fun () -> NegEqOR(ftrma_trm name_tp [] m1,ftrma_trm name_tp [] m2,r1 (),r2 ()))))
		     end
		  | Unsatisfiable(None) ->
		     reset_search_1 (); incr sgn;
		     split_global_disjunctions ((fneg m1)::m2::pr) b2 b bs sgn cnj
		  | Timeout ->
		     if (!nontheorem) then
		       begin
			 let s = get_timeout_default 0.0 in
			 if (s >= 0.2) then
			   begin
			     set_timer (s *. 0.5);
			     mult_timeout 0.5;
			     reset_search_1 (); incr sgn;
			     split_global_disjunctions ((fneg m1)::m2::pr) b2 b bs sgn cnj
			   end
			 else if (s >= 0.0) then
			   begin
			     set_timer s;
			     timeout := Some 0.0;
			     reset_search_1 (); incr sgn;
			     split_global_disjunctions ((fneg m1)::m2::pr) b2 b bs sgn cnj
			   end
			 else raise Timeout
		       end
		     else
		       raise Timeout
		end
             | _ -> split_global_disjunctions pr b2 (m::b) bs sgn cnj
           end
	| _ -> split_global_disjunctions pr b2 (m::b) bs sgn cnj
      end
  | [] ->
      begin
	match b2 with
	  (_::_) -> split_global_disjunctions b2 [] [] b sgn false
	| [] ->
	    begin
	      if (!verbosity > 5) then Printf.printf "Searching on Subgoal %d\n" (!sgn);
	      if cnj then
		search_1 b bs
	      else
		search_1 bs b (*** reversed conjecture and axiom parts to compute axiom parts ***)
	    end
      end

let search () =
  minisatsearchperiod := get_int_flag "MINISAT_SEARCH_PERIOD";
  useeprover := get_bool_flag "USE_E";
  eeqoequiv := get_bool_flag "E_EQO_EQUIV";
  eproverperiod := get_int_flag "E_PERIOD";
  eprovertimeout := get_int_flag "E_TIMEOUT";
(*  if (get_bool_flag "ENABLE_PATTERN_CLAUSES") then
    begin
      enable_pattern_clauses := true;
      if (get_bool_flag "DYNAMIC_PATTERN_CLAUSES") then
	dynamic_pattern_clauses := true;
      if (get_bool_flag "PATTERN_CLAUSES_EARLY") then
	begin
	  get_literal_fn1 := get_literal_and_make_pattern_clause;
	  if (!dynamic_pattern_clauses) then
	    begin
	      get_literal_fn2 := get_literal_and_make_pattern_clause;
	      dynamic_pattern_clauses := false
	    end
	end
    end; *)
  begin
    if (not (!nontheorem)) then
      begin
        if (get_bool_flag "BASETYPESTOPROP") then
           raise (GenericError "The flag BASETYPESTOPROP must be false unless trying to determine (Counter)Satisfiablity (-N on the command line).");
        if (get_bool_flag "BASETYPESFINITE") then
           raise (GenericError "The flag BASETYPESFINITE must be false unless trying to determine (Counter)Satisfiablity (-N on the command line).");
      end;
    search_init();
    pattern_clauses_forall_as_lit := get_bool_flag "PATTERN_CLAUSES_FORALL_AS_LIT"; (*** Chad: April 6, 2011 ***)
    pattern_clauses_onlyallstrict := get_bool_flag "PATTERN_CLAUSES_ONLYALLSTRICT"; (*** Chad: April 6, 2011 ***)
    pattern_clauses_eqnlits := get_bool_flag "PATTERN_CLAUSES_EQNLITS"; (*** Chad: April 6, 2011 ***)
    begin (*** Chad: Nov 2011 ***)
      not_in_prop_model_delay := get_int_flag "NOT_IN_PROP_MODEL_DELAY";
      not_in_prop_model_delay_p := (!not_in_prop_model_delay > 0)
    end;
    let b1 = ref [] in
    let b2 = ref [] in
    if (get_bool_flag "TREAT_CONJECTURE_AS_SPECIAL") then
      ignore (List.map (fun ((_,_,_,mh) as m) -> if (Hashtbl.mem part_of_conjecture mh) then b1 := (m::!b1) else b2 := (m::!b2)) (!initial_branch)) (*** March 18 2011; split branch into conjecture part (b1) and axiom part (b2) ***)
    else
      ignore (b2 := (!initial_branch)); (*** March 20 2011: In this case it should behave essentially as Satallax 2.0 did. ***)
    if (get_bool_flag "SPLIT_GLOBAL_DISJUNCTIONS") then
      begin
	if (!nontheorem) then
	  begin
	    match (!timeout) with
	    | Some tm -> nontheorem_split_timeout := tm
	    | None -> ()
	  end;
	if (get_bool_flag "PREPROCESS_BEFORE_SPLIT") then
	  let pbr1 = (List.map preprocess1 (!b1)) in
	  let pbr2 = (List.map preprocess1 (!b2)) in
	  already_preprocessed := true;
	  split_global_disjunctions pbr1 pbr2 [] [] (ref 1) true (*** Following Andreas Teucke's suggestion: Preprocess before split - Mar 18 2011 ***)
	else
	  split_global_disjunctions (!b1) (!b2) [] [] (ref 1) true
      end
    else
      search_1 (!b1) (!b2)
  end

